#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#include "stack.h"


static int verbose = 0 ;

/* Popper */
typedef struct {
  stack_t *stack;
  volatile int sum;
} common_popper_t ;

common_popper_t *alloc_common_popper(stack_t *stack) {
  common_popper_t *p = malloc_check(sizeof(*p)) ;
  p->stack = stack ;
  p->sum = 0 ;
  return p ;
}

void free_common_popper(common_popper_t *p) {
  free(p);
}

typedef struct {
  int id ;
  common_popper_t *common ;
} popper_t ;

popper_t *alloc_popper_t(int id,common_popper_t *c) {
  popper_t *p = malloc_check(sizeof(*p)) ;
  p->id = id ;
  p->common = c ;
  return p ;
}

void free_popper_t(popper_t *p) { free(p); }

void *popper(void *p) {
  popper_t *w = (popper_t *)p ;
  common_popper_t *c = w->common ;
  int id = w->id ;
  free_popper_t(w) ;
  void *item ;
  while ((item = pop(c->stack)) != NULL) {
    boxed_int_t *q = item ;
    int v = q->v ;
    free_boxed_int(q) ;
    (void)__sync_fetch_and_add(&c->sum,v);
    if (verbose) fprintf(stderr,"POPPER<%i> GOT %i\n",id,v) ;
  }
  if (verbose) fprintf(stderr,"POPPER<%i> OUT\n",id) ;
  return NULL ;
}


/* Create n poppers */
common_popper_t *spawn_poppers(stack_t *stack, int n) {
  common_popper_t *c = alloc_common_popper(stack);
  for (int id = 0 ; id < n ; id++) {
    popper_t *w = alloc_popper_t(id,c) ;
    create_thread_detached(popper,w) ;
  }
  return c ;
}

/* Pushers */

typedef struct {
  stack_t *stack;
} common_pusher_t ;

common_pusher_t *alloc_common_pusher(stack_t *stack) {
  common_pusher_t *p = malloc_check(sizeof(*p)) ;
  p->stack = stack ;
  return p ;
}

void free_common_pusher(common_pusher_t *p) {
  free(p);
}

typedef struct {
  int id ;
  common_pusher_t *common ;
} pusher_t ;

pusher_t *alloc_pusher_t(int id,common_pusher_t *c) {
  pusher_t *p = malloc_check(sizeof(*p)) ;
  p->id = id ;
  p->common = c ;
  return p ;
}

void free_pusher_t(pusher_t *p) { free(p); }

void *pusher(void *p) {
  pusher_t *w = (pusher_t *)p ;
  common_pusher_t *c = w->common ;
  int id = w->id ;
  free_pusher_t(w) ;
  // push 1 id+1 times
  for (int k = 0 ; k <= id ; k++) {
    if (verbose) fprintf(stderr,"PUSHER<%i> PUT %i\n",id,1) ;
    push(c->stack,alloc_boxed_int(1)) ;
  }
  return NULL ;
}

/* Create n pushers */
common_pusher_t *spawn_pushers(stack_t *stack, int n) {
  common_pusher_t *c = alloc_common_pusher(stack);
  pthread_t th[n] ;
  for (int id = 0 ; id < n ; id++) {
    pusher_t *w = alloc_pusher_t(id,c) ;
    create_thread(&th[id],pusher,w) ;
  }
  for (int id = 0 ; id < n ; id++) join_thread(&th[id]);
  return c ;
}

/* Main driver */

void zyva(int nprocs,int sz) {
  if (verbose) fprintf(stderr,"nprocs=%i, sz=%i\n",nprocs,sz);
  stack_t *stack = alloc_stack(sz) ;
  common_popper_t *pop = spawn_poppers(stack,nprocs) ;
  common_pusher_t *push = spawn_pushers(stack,nprocs) ;
  kill(stack,nprocs) ;
  int sum = __sync_fetch_and_add(&pop->sum,0) ;
  int ok = 0 ;
  for (int k = 1 ; k <= nprocs ; k++) ok += k ;
  printf("SUM=%i, OK=%i\n",sum,ok);
  free_common_popper(pop) ;
  free_common_pusher(push) ;
  free_stack(stack) ;
}

/* Main function */

void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [nprocs] [sz]\n",pgm) ;
}
int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int nprocs = 2 ;
  int sz = 100 ;
  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    nprocs = atoi(argv[0]) ;
  } else if (argc == 2) {
    nprocs = atoi(argv[0]) ;
    sz = atoi(argv[1]) ;
  }
  zyva(nprocs,sz) ;
  return 0 ;
}
